import os
import shutil
import libvirt
import subprocess

from . import generatename
from . import progress
from .guest import Guest
from .devices import DeviceDisk
from .logger import log


def _generate_backup_disk_path(conn, origname, newname, origpath):
    """
    Generate desired backuped disk path name, derived from the
    original path, original VM name, and proposed new VM name
    """
    if origpath is None:
        return None

    path = origpath
    suffix = ""

    # Try to split the suffix off the existing disk name. Ex.
    # foobar.img -> foobar-backup.img
    #
    # If the suffix is greater than 7 characters, assume it isn't
    # a file extension and is part of the disk name, at which point
    # just stick '-backup' on the end.
    if "." in origpath and len(origpath.rsplit(".", 1)[1]) <= 7:
        path, suffix = origpath.rsplit(".", 1)
        suffix = "." + suffix

    dirname = os.path.dirname(path)
    basename = os.path.basename(path)

    backupbase = basename + "-backup"
    if origname and basename == origname:
        backupbase = newname

    backupbase = os.path.join(dirname, backupbase)
    def cb(p):
        return DeviceDisk.path_definitely_exists(conn, p)
    return generatename.generate_name(backupbase, cb, suffix=suffix)


def _lookup_vm(conn, name):
    try:
        return conn.lookupByName(name)
    except libvirt.libvirtError:
        e = ValueError(_("Domain '%s' was not found.") % str(name))
        raise e from None


def _get_backupable_msg(disk):
    """
    If the disk storage is not backupable, return a string explaining why
    """
    if disk.wants_storage_creation():
        return _("Disk path '%s' does not exist.") % disk.get_source_path()

    if disk.type == "network":
        proto = disk.source.protocol
        if proto == "rbd":
            # This case, rbd with managed storage, is implementable. It
            # requires open coding a bunch of work in cloner, or reworking
            # other disk code to add unique URIs for rbd volumes and pools
            return (
                _("Cloning rbd volumes is not yet supported.") +
                " https://github.com/virt-manager/virt-manager/issues/177")
        return _("Disk network type '%s' is not backupable.") % proto


def _get_shareable_msg(disk):
    if disk.is_empty():
        return _("No storage to backup.")
    if disk.read_only:
        return _("Read Only")
    if disk.shareable or disk.transient_shareBacking:
        return _("Marked as shareable")


class _CloneDiskInfo:
    """
    Class that tracks some additional information about how we want
    to default handle each disk of the source VM

    For any source disk there's 3 main scenarios:

    * clone: Copy contents from src to dst. If dst path doesn't
            exist we attempt to create it. If it exists we overwrite it
    * preserve: Destination path is an existing, and no copying is performed.
    * share: Original disk XML is used unchanged for the new disk
    """
    _ACTION_SHARE = 1
    _ACTION_CLONE = 2
    _ACTION_PRESERVE = 3

    def __init__(self, srcdisk):
        self.disk = DeviceDisk(srcdisk.conn, parsexml=srcdisk.get_xml())
        self.disk.set_backend_for_existing_path()
        self.new_disk = None

        self._share_msg = _get_shareable_msg(self.disk)
        self._cloneable_msg = -1
        self._newpath_msg = None

        self._action = None
        self.set_clone_requested()
        if self.get_share_msg():
            self.set_share_requested()

    def is_clone_requested(self):
        return self._action in [self._ACTION_CLONE]
    def is_share_requested(self):
        return self._action in [self._ACTION_SHARE]
    def is_preserve_requested(self):
        return self._action in [self._ACTION_PRESERVE]

    def _set_action(self, action):
        if action != self._action:
            self._action = action
    def set_clone_requested(self):
        self._set_action(self._ACTION_CLONE)
    def set_share_requested(self):
        self._set_action(self._ACTION_SHARE)
    def set_preserve_requested(self):
        self._set_action(self._ACTION_PRESERVE)

    def get_share_msg(self):
        return self._share_msg
    def get_exporable_msg(self):
        if self._cloneable_msg == -1:
            self._cloneable_msg = _get_backupable_msg(self.disk)
        return self._cloneable_msg
    def get_newpath_msg(self):
        return self._newpath_msg

    def raise_error(self):
        if self.is_clone_requested() and self.get_exporable_msg():
            msg = self.get_exporable_msg()
            err = _("Could not determine original disk information: %s" % msg)
            raise ValueError(err)
        if self.is_share_requested():
            return
        if self.get_newpath_msg():
            msg = self.get_newpath_msg()
            raise ValueError(msg)


class Backuper(object):
    @staticmethod
    def generate_backup_disk_path(conn, origname, newname, origpath):
        return _generate_backup_disk_path(conn, origname, newname, origpath)

    def __init__(self, conn, src_name=None, src_xml=None):
        self.conn = conn

        self._src_guest = None
        self._new_name = None
        self._diskinfos = []
        self._nvram_diskinfo = None
        self._init_src(src_name, src_xml)

    #################
    # Init routines #
    #################

    def _init_src(self, src_name, src_xml):
        """
        Set up the source VM info we are backuping, from passed in VM name
        or full XML
        """
        if not src_xml:
            dom = _lookup_vm(self.conn, src_name)
            flags = libvirt.VIR_DOMAIN_XML_SECURE
            src_xml = dom.XMLDesc(flags)

        self._src_guest = Guest(self.conn, parsexml=src_xml)

        # Collect disk info for every disk to determine if we will
        # default to backuping or not
        for disk in self._src_guest.devices.disk:
            self._diskinfos.append(_CloneDiskInfo(disk))
        for diskinfo in [d for d in self._diskinfos if d.is_clone_requested()]:
            disk = diskinfo.disk
            log.debug("Wants backuping: size=%s path=%s",
                    disk.get_size(), disk.get_source_path())


    ##############
    # Properties #
    ##############

    @property
    def src_name(self):
        """
        The name of the original VM we are exporting
        """
        return self._src_guest.name

    @property
    def nvram_diskinfo(self):
        return self._nvram_diskinfo

    def get_diskinfos(self):
        """
        Return the list of _CloneDiskInfo instances
        """
        return self._diskinfos[:]

    def get_nonshare_diskinfos(self):
        """
        Return a list of _CloneDiskInfo that are tagged for exporting
        """
        return [di for di in self.get_diskinfos() if
                not di.is_share_requested()]


    ######################
    # Functional methods #
    ######################

    def start_backuping(self, cron, command, create_cron, meter=None):
        """
        Actually perform the backuping
        """
        log.debug("Starting backuping.")
        meter = progress.ensure_meter(meter)
        backup_path = command.split("-o ")[1].split(" --threshold")[0]
        backup_exists = os.path.exists(backup_path)
        try:
            meter.start(_("Processing..."), size=0)
            result = str(subprocess.run(command, shell=True,
                                    capture_output=True, text=True))
            # log.debug(result)
            if "Permission denied" in result:
                if not backup_exists:
                    shutil.rmtree(backup_path)
                raise Exception(_("Permission denied (try starting the virtual "
                                  "machine)."))
            if "ERROR" in result:
                error_message = result
                if "[MainThread]: " in result:
                    error_message = result.rsplit("[MainThread]: ", 1)[1][:-4]
                raise Exception(error_message)
            if create_cron:
                cron.write()
        except:
            raise
        finally:
            meter.end()

        log.debug("Backuping finished.")
