import re
import os
import shutil
import tarfile
import tempfile
import hashlib
import libvirt

from . import generatename
from . import progress
from . import xmlutil
from .guest import Guest
from .devices import DeviceInterface
from .devices import DeviceDisk
from .logger import log
from .devices import DeviceChannel


def _download_disk(vol, st, new_disk_file, meter, size):
    def recvSkipHandler(stream, length, opaque):
        fd = opaque
        cur = os.lseek(fd, length, os.SEEK_CUR)
        return os.ftruncate(fd, cur)

    def bytesWriteHandler(stream, buf, opaque):
        nonlocal size
        size += len(buf)
        meter.update(size)
        fd = opaque
        return os.write(fd, buf)

    offset = 0
    length = 0
    fd = os.open(new_disk_file, os.O_WRONLY | os.O_CREAT | os.O_TRUNC,
                mode=0o0660)
    vol.download(st, offset, length,
                libvirt.VIR_STORAGE_VOL_DOWNLOAD_SPARSE_STREAM)
    st.sparseRecvAll(bytesWriteHandler, recvSkipHandler, fd)

    os.close(fd)

    return size

def _replace_vm(conn, name):
    """
    Remove the existing VM with the same name if requested
    """
    try:
        vm = conn.lookupByName(name)
    except libvirt.libvirtError:
        return

    try:

        log.debug("Explicitly replacing guest '%s'", name)
        if vm.ID() != -1:
            log.debug("Destroying guest '%s'", name)
            vm.destroy()

        log.debug("Undefining guest '%s'", name)
        vm.undefine()
    except libvirt.libvirtError as e:  # pragma: no cover
        msg = (_("Could not remove old vm '%(vm)s': %(error)s") % {
                "vm": name, "error": str(e)})
        raise RuntimeError(msg) from None


def _generate_export_name(conn, basename):
    """
    If the orig name is "foo-export", we don't want the export to be
    "foo-export-export", we want "foo-export1"
    """
    regex = r"-export[1-9]*$"
    match = re.search(regex, basename)
    start_num = 1
    force_num = False
    if match:
        num_match = re.search("[1-9]+$", match.group())
        force_num = True
        if num_match:
            start_num = int(str(num_match.group())) + 1
        basename = basename[:match.start()]

    def cb(n):
        return generatename.check_libvirt_collision(
            conn.lookupByName, n)
    basename = basename + "-export"
    return generatename.generate_name(basename, cb,
            sep="", start_num=start_num, force_num=force_num)


def _generate_export_disk_path(conn, origname, newname, origpath):
    """
    Generate desired exported disk path name, derived from the
    original path, original VM name, and proposed new VM name
    """
    if origpath is None:
        return None

    path = origpath
    suffix = ""

    # Try to split the suffix off the existing disk name. Ex.
    # foobar.img -> foobar-export.img
    #
    # If the suffix is greater than 7 characters, assume it isn't
    # a file extension and is part of the disk name, at which point
    # just stick '-export' on the end.
    if "." in origpath and len(origpath.rsplit(".", 1)[1]) <= 7:
        path, suffix = origpath.rsplit(".", 1)
        suffix = "." + suffix

    dirname = os.path.dirname(path)
    basename = os.path.basename(path)

    exportbase = basename + "-export"
    if origname and basename == origname:
        exportbase = newname

    exportbase = os.path.join(dirname, exportbase)
    def cb(p):
        return DeviceDisk.path_definitely_exists(conn, p)
    return generatename.generate_name(exportbase, cb, suffix=suffix)


def _lookup_vm(conn, name):
    try:
        return conn.lookupByName(name)
    except libvirt.libvirtError:
        e = ValueError(_("Domain '%s' was not found.") % str(name))
        raise e from None


def _build_clone_vol_install(orig_disk, new_disk):
    # We set a stub size for initial creation
    # set_input_vol will overwrite it
    size = .000001
    sparse = False
    vol_install = DeviceDisk.build_vol_install(
        orig_disk.conn, os.path.basename(new_disk.get_source_path()),
        new_disk.get_parent_pool(), size, sparse)
    vol_install.set_input_vol(orig_disk.get_vol_object())

    return vol_install


def _build_export_disk(orig_disk, exportpath, allow_create, sparse):
    conn = orig_disk.conn
    device = DeviceDisk.DEVICE_DISK
    if not exportpath:
        device = DeviceDisk.DEVICE_CDROM

    new_disk = DeviceDisk(conn)
    new_disk.set_source_path(exportpath)
    new_disk.device = device

    if not allow_create:
        new_disk.validate()
        return new_disk

    if new_disk.get_vol_object():
        # Special case: non remote  of a guest using
        # managed block devices: fall back to local cloning if
        # we have permissions to do so. This validation check
        # caused a few bug reports in a short period of time,
        # so must be a common case.
        if (conn.is_remote() or
            new_disk.type != new_disk.TYPE_BLOCK or
            not orig_disk.get_source_path() or
            not os.access(orig_disk.get_source_path(), os.R_OK) or
            not new_disk.get_source_path() or
            not os.access(new_disk.get_source_path(), os.W_OK)):
            raise RuntimeError(
                _("Clone onto existing storage volume is not "
                  "currently supported: '%s'") % new_disk.get_source_path())

    if (orig_disk.get_vol_object() and
        new_disk.wants_storage_creation()):
        vol_install = _build_clone_vol_install(orig_disk, new_disk)
        if not sparse:
            vol_install.allocation = vol_install.capacity
        new_disk.set_vol_install(vol_install)
    elif orig_disk.get_source_path():
        new_disk.set_local_disk_to_clone(orig_disk, sparse)

    new_disk.validate()
    return new_disk


def _get_exportable_msg(disk):
    """
    If the disk storage is not exportable, return a string explaining why
    """
    if disk.wants_storage_creation():
        return _("Disk path '%s' does not exist.") % disk.get_source_path()

    if disk.type == "network":
        proto = disk.source.protocol
        if proto == "rbd":
            # This case, rbd with managed storage, is implementable. It
            # requires open coding a bunch of work in cloner, or reworking
            # other disk code to add unique URIs for rbd volumes and pools
            return (
                _("Cloning rbd volumes is not yet supported.") +
                " https://github.com/virt-manager/virt-manager/issues/177")
        return _("Disk network type '%s' is not exportable.") % proto


def _get_shareable_msg(disk):
    if disk.is_empty():
        return _("No storage to clone.")
    if disk.read_only:
        return _("Read Only")
    if disk.shareable or disk.transient_shareBacking:
        return _("Marked as shareable")


class _CloneDiskInfo:
    """
    Class that tracks some additional information about how we want
    to default handle each disk of the source VM

    For any source disk there's 3 main scenarios:

    * clone: Copy contents from src to dst. If dst path doesn't
            exist we attempt to create it. If it exists we overwrite it
    * preserve: Destination path is an existing, and no copying is performed.
    * share: Original disk XML is used unchanged for the new disk
    """
    _ACTION_SHARE = 1
    _ACTION_CLONE = 2
    _ACTION_PRESERVE = 3

    def __init__(self, srcdisk):
        self.disk = DeviceDisk(srcdisk.conn, parsexml=srcdisk.get_xml())
        self.disk.set_backend_for_existing_path()
        self.new_disk = None

        self._share_msg = _get_shareable_msg(self.disk)
        self._cloneable_msg = -1
        self._newpath_msg = None

        self._action = None
        self.set_clone_requested()
        if self.get_share_msg():
            self.set_share_requested()

    def is_clone_requested(self):
        return self._action in [self._ACTION_CLONE]
    def is_share_requested(self):
        return self._action in [self._ACTION_SHARE]
    def is_preserve_requested(self):
        return self._action in [self._ACTION_PRESERVE]

    def _set_action(self, action):
        if action != self._action:
            self._action = action
    def set_clone_requested(self):
        self._set_action(self._ACTION_CLONE)
    def set_share_requested(self):
        self._set_action(self._ACTION_SHARE)
    def set_preserve_requested(self):
        self._set_action(self._ACTION_PRESERVE)

    def set_new_path(self, path, sparse):
        allow_create = not self.is_preserve_requested()
        if allow_create:
            msg = self.get_exporable_msg()
            if msg:
                return

        try:
            self.new_disk = Exporter.build_export_disk(
                    self.disk, path, allow_create, sparse)
        except Exception as e:
            log.debug("Error setting clone path.", exc_info=True)
            err = (_("Could not use path '%(path)s' for exporting: %(error)s") %
                    {"path": path, "error": str(e)})
            self._newpath_msg = err

    def get_share_msg(self):
        return self._share_msg
    def get_exporable_msg(self):
        if self._cloneable_msg == -1:
            self._cloneable_msg = _get_exportable_msg(self.disk)
        return self._cloneable_msg
    def get_newpath_msg(self):
        return self._newpath_msg

    def raise_error(self):
        if self.is_clone_requested() and self.get_exporable_msg():
            msg = self.get_exporable_msg()
            err = _("Could not determine original disk information: %s" % msg)
            raise ValueError(err)
        if self.is_share_requested():
            return
        if self.get_newpath_msg():
            msg = self.get_newpath_msg()
            raise ValueError(msg)


class Exporter(object):
    @staticmethod
    def generate_export_name(conn, basename):
        return _generate_export_name(conn, basename)

    @staticmethod
    def generate_export_disk_path(conn, origname, newname, origpath):
        return _generate_export_disk_path(conn, origname, newname, origpath)

    @staticmethod
    def build_export_disk(orig_disk, exportpath, allow_create, sparse):
        return _build_export_disk(orig_disk, exportpath, allow_create, sparse)

    def __init__(self, conn, src_name=None, src_xml=None):
        self.conn = conn

        self._src_guest = None
        self._new_guest = None
        self._diskinfos = []
        self._nvram_diskinfo = None
        self._init_src(src_name, src_xml)

        self._new_nvram_path = None

        self._sparse = True
        self._replace = False
        self._reflink = False


    #################
    # Init routines #
    #################

    def _init_src(self, src_name, src_xml):
        """
        Set up the source VM info we are exporting, from passed in VM name
        or full XML
        """
        if not src_xml:
            dom = _lookup_vm(self.conn, src_name)
            status = dom.info()[0]
            if status not in [libvirt.VIR_DOMAIN_SHUTOFF]:
                raise RuntimeError(_("Domain to export must be shutoff."))
            flags = libvirt.VIR_DOMAIN_XML_SECURE
            src_xml = dom.XMLDesc(flags)

        self._src_guest = Guest(self.conn, parsexml=src_xml)
        self._new_guest = Guest(self.conn, parsexml=src_xml)
        self._init_new_guest()

        # Collect disk info for every disk to determine if we will
        # default to exporting or not
        for disk in self._src_guest.devices.disk:
            self._diskinfos.append(_CloneDiskInfo(disk))
        for diskinfo in [d for d in self._diskinfos if d.is_clone_requested()]:
            disk = diskinfo.disk
            log.debug("Wants exporing: size=%s path=%s",
                    disk.get_size(), disk.get_source_path())

        if self._src_guest.os.nvram:
            old_nvram = DeviceDisk(self.conn)
            old_nvram.set_source_path(self._new_guest.os.nvram)
            self._nvram_diskinfo = _CloneDiskInfo(old_nvram)

    def _init_new_guest(self):
        """
        Perform the series of unconditional new VM changes we always make
        """
        self._new_guest.id = None
        self._new_guest.title = None
        self._new_guest.uuid = None
        self._new_guest.set_backup_schedule(None)
        self._new_guest.set_backup_repeat(None)
        self._new_guest.set_backup_folder(None)
        self._new_guest.set_backup_type(None)
        self._new_guest.set_backup_threshold(None)
        self._new_guest.set_backup_excluded(None)
        self._new_guest.set_backup_compress(None)
        self._new_guest.set_backup_included(None)

        for dev in self._new_guest.devices.graphics:
            if dev.port and dev.port != -1:
                log.warning(_("Setting the graphics device port to autoport, "
                               "in order to avoid conflicting."))
                dev.port = -1

        for iface in self._new_guest.devices.interface:
            iface.target_dev = None
            iface.macaddr = None

        # For guest agent channel, remove a path to generate a new one with
        # new guest name
        for channel in self._new_guest.devices.channel:
            if (channel.type == DeviceChannel.TYPE_UNIX and
                channel.target_name and channel.source.path and
                channel.target_name in channel.source.path):
                channel.source.path = None

        new_name = Exporter.generate_export_name(self.conn, self.src_name)
        log.debug("Auto-generated export name '%s'", new_name)
        self.set_export_name(new_name)


    ##############
    # Properties #
    ##############

    @property
    def src_name(self):
        """
        The name of the original VM we are exporting
        """
        return self._src_guest.name

    @property
    def new_guest(self):
        """
        The Guest instance of the new XML we will create
        """
        return self._new_guest

    @property
    def nvram_diskinfo(self):
        return self._nvram_diskinfo

    def set_export_name(self, name):
        self._new_guest.name = name

    def set_export_uuid(self, uuid):
        """
        Override the new VMs generated UUId
        """
        self._new_guest.uuid = uuid

    def set_replace(self, val):
        """
        If True, don't check for export name collision, simply undefine
        any conflicting guest.
        """
        self._replace = bool(val)

    def set_reflink(self, reflink):
        """
        If true, use COW lightweight copy
        """
        self._reflink = reflink

    def set_sparse(self, flg):
        """
        If True, attempt sparse allocation during exporting
        """
        self._sparse = flg

    def get_diskinfos(self):
        """
        Return the list of _CloneDiskInfo instances
        """
        return self._diskinfos[:]

    def get_nonshare_diskinfos(self):
        """
        Return a list of _CloneDiskInfo that are tagged for exporting
        """
        return [di for di in self.get_diskinfos() if
                not di.is_share_requested()]

    def get_share_diskinfos(self):
        """
        Return a list of _CloneDiskInfo that are tagged for not exporting
        """
        return [di for di in self.get_diskinfos() if di.is_share_requested()]

    def set_nvram_path(self, val):
        """
        If the VM needs to have nvram content cloned, this overrides the
        destination path
        """
        self._new_nvram_path = val


    ######################
    # Functional methods #
    ######################

    def _prepare_nvram(self):
        if not self._nvram_diskinfo:
            return

        new_nvram_path = self._new_nvram_path
        if new_nvram_path is None:
            nvram_dir = os.path.dirname(self._new_guest.os.nvram)
            new_nvram_path = os.path.join(
                    nvram_dir, "%s_VARS.fd" % self._new_guest.name)

        diskinfo = self._nvram_diskinfo
        new_nvram = DeviceDisk(self.conn)
        new_nvram.set_source_path(new_nvram_path)
        old_nvram = DeviceDisk(self.conn)
        old_nvram.set_source_path(diskinfo.disk.get_source_path())

        if (diskinfo.is_clone_requested() and
            new_nvram.wants_storage_creation() and
            diskinfo.disk.get_vol_object()):
            # We only run validation if there's some existing nvram we
            # can copy. It's valid for nvram to not exist at VM define
            # time, libvirt will create it for us
            diskinfo.set_new_path(new_nvram_path, self._sparse)
            diskinfo.raise_error()
            diskinfo.new_disk.get_vol_install().reflink = self._reflink
        else:
            # There's no action to perform for this case, so drop it
            self._nvram_diskinfo = None

        self._new_guest.os.nvram = new_nvram.get_source_path()


    def prepare(self):
        """
        Validate and set up all parameters needed for the new (export) VM
        """
        try:
            Guest.validate_name(self.conn, self._new_guest.name,
                                check_collision=not self._replace,
                                validate=False)
        except ValueError as e:
            msg = _("Invalid name for new guest: %s") % e
            raise ValueError(msg) from None

        for diskinfo in self.get_nonshare_diskinfos():
            orig_disk = diskinfo.disk

            if not diskinfo.new_disk:
                # User didn't set a path, generate one
                newpath = Exporter.generate_export_disk_path(
                         self.conn, self.src_name,
                         self.new_guest.name,
                         orig_disk.get_source_path())
                diskinfo.set_new_path(newpath, self._sparse)
                if not diskinfo.new_disk:
                    # We hit an error, clients will raise it later
                    continue

            new_disk = diskinfo.new_disk
            assert new_disk
            log.debug("Exporting srcpath=%s dstpath=%s",
                    orig_disk.get_source_path(), new_disk.get_source_path())

            if self._reflink:
                vol_install = new_disk.get_vol_install()
                vol_install.reflink = self._reflink

            for disk in self._new_guest.devices.disk:
                if disk.target == orig_disk.target:
                    xmldisk = disk

            # Change the XML
            xmldisk.set_source_path(None)
            xmldisk.type = new_disk.type
            xmldisk.driver_name = orig_disk.driver_name
            xmldisk.driver_type = orig_disk.driver_type
            xmldisk.set_source_path(new_disk.get_source_path())

        self._prepare_nvram()

        # Save altered export xml
        diff = xmlutil.diff(self._src_guest.get_xml(),
                self._new_guest.get_xml())
        log.debug("Export guest xml diff:\n%s", diff)
                 

    def start_exporting(self, targz_name, snapshots, compress_level,
                        checksum, meter=None):
        """
        Actually perform the exporting: copying disks if needed and defining
        the new export xml.
        """
        log.debug("Starting exporting.")
        meter = progress.ensure_meter(meter)

        tmp = tempfile.mkdtemp()
        try:
            # Disks
            diskinfos = self.get_diskinfos()
            if self._nvram_diskinfo:
                diskinfos.append(self._nvram_diskinfo)

            disks_size = self.get_disks_size(diskinfos)
            meter.start("Copying disks...", disks_size)

            self.save_disks(diskinfos, tmp, meter)
            self.save_loader(tmp)
            self.save_xml(tmp)
            self.save_snapshots_xml(snapshots, tmp)

            meter.end()
            meter.start("Archiving...", size=0)
            self.archive_folder(targz_name, tmp, compress_level)

            if checksum:
                self.create_checksum(targz_name)

        except Exception as e:
            log.debug("Exporting failed: %s", str(e))
            raise

        finally:
            meter.end()
            shutil.rmtree(tmp)

        log.debug("Exporting finished.")

    def get_disks_size(self, diskinfos):
        size = 0
        for diskinfo in diskinfos:
            if not diskinfo.is_clone_requested():
                continue
            size += int(diskinfo.disk.get_size() * 1024 * 1024 * 1024)
        return size
    
    def save_disks(self, diskinfos, tmp, meter):
        size = 0
        for diskinfo in diskinfos:
            if not diskinfo.is_clone_requested():
                continue
            
            st = self.conn.newStream()
            vol = self.conn.storageVolLookupByPath(diskinfo.disk
                                                    .get_source_path())

            new_disk_file = os.path.join(tmp, diskinfo.new_disk
                                            .get_vol_install().name)
            size += _download_disk(vol, st, new_disk_file, meter, size)
            st.finish()
    
    def save_loader(self, tmp):
        try:
            loader = self._src_guest.os.loader
            if loader:
                shutil.copy(loader, tmp)
        except:
            pass

    def save_xml(self, tmp):
        xml_path = os.path.join(tmp, f"{self._new_guest.name}.xml")
        with open(xml_path, 'w') as f:
            f.write(self._new_guest.get_xml())
    
    def save_snapshots_xml(self, snapshots, tmp):
        snap_xml_folder = os.path.join(tmp, "snapshots")
        os.makedirs(snap_xml_folder, exist_ok=True)

        for snap in snapshots:
            snap_xml_file = f"{snap.get_name()}.xml"
            snap_xml_path = os.path.join(snap_xml_folder, snap_xml_file)
            with open(snap_xml_path, 'w') as f:
                f.write(snap.xmlobj.get_xml())
    
    def archive_folder(self, targz_name, tmp, compresslevel):
        with tarfile.open(targz_name, "w:gz",
                          compresslevel=compresslevel) as tar:
            for file_name in os.listdir(tmp):
                file_path = os.path.join(tmp, file_name)
                tar.add(file_path, arcname=os.path.relpath(file_path, tmp))
        tar.close()

    def create_checksum(self, targz_name):
        md5 = hashlib.md5()

        with open(targz_name, 'rb') as f:
            while True:
                chunk = f.read(4096)
                if not chunk:
                    break
                md5.update(chunk)

        with open(f"{targz_name}.md5", 'w') as f:
            f.write(md5.hexdigest())
